#! /usr/bin/env python3
# encoding: utf-8

from collections import OrderedDict
import enum
import sys
from . import docs
from . import text
from . import types

# --------------------------------------------------

class Option_sources(enum.Enum):
	none = 0
	default = 1
	config = 2
	cmdline = 3

class Option:
	def __init__(self, name, description, callback, default=Option_sources.none, count=None, abbr=None, level=2, in_cfg_file=True, in_help=True):
		self.name = name
		self.description = description
		self.callback = callback
		self.default = default
		self.count = count
		self.abbr = abbr
		self.level = level
		self.in_cfg_file = in_cfg_file
		self.in_help = in_help
		
		if default == Option_sources.none:
			self.source = Option_sources.none
			self._value = Option_sources.none
		else:
			self.source = Option_sources.default
			self._value = default
	
	def set_value(self, value, source):
		self._value = value
		self.source = source
	
	@property
	def value(self):
		if self.source == Option_sources.none:
			raise AttributeError("option has no value")
		else:
			return self._value
	
	def dump(self):
		if self.source == Option_sources.none:
			return None
		else:
			if self.callback == bool:
				if self.value:
					...

# --------------------------------------------------

class ConfigRouter:
	def __init__(self, options):
		self.options = options
	
	def __getattr__(self, name):
		return self.options[name].value

# --------------------------------------------------

class Main:
	"""
	Application backbone. Provides:
	  * A configuration system consisting of
	    * a given set of (hardcoded) default values
	    * a configuration file that overrides the defaults
	    * a command line parser that overrides all of the above
	  * Help system that generates help text based on the application's configurable options.
	"""
	
	def __init__(self, name, version=None, license=None, use_cfg_file=False, auto_add_help=True):
		self.name = name
		self.version = version
		self.license = license
		self.print = text.Output(name)
		self.options = OrderedDict() # level: ndict
		self.cfg = ConfigRouter(self.options)
		
		if use_cfg_file:
			...
		
		if auto_add_help:
			self.enable_help("help", "h")
	
	def __repr__(self):
		return 'over.app.Main(name="%s")' %(self.name)
	
	def exit(self, rv=0):
		"""
		Terminates the program and returns `rv`.
		"""
		
		sys.exit(rv)
	
	def add_option(self, *args, **kwargs):
		option = Option(*args, **kwargs)
		
		if option.level not in self.options:
			self.options[option.level] = types.ndict()
		
		self.options[option.level][option.name] = option
	
	def add_doc(self, chapter, paragraphs):
		...
	
	def enable_help(self, name, abbr):
		"""
		Map application help to --name and -abbr, and enable library ---help.
		"""
		
		self.add_option("help", "Display the application configuration view.", self.help, abbr=abbr, in_cfg_file=False)
		self.add_option("help", "Display the library help and about views.", lambda: self.help(docs=docs.over_desc), level=3, in_cfg_file=False)
	
	def help(self, option=None, docs=None):
		"""
		Displays a help text and exits the program.
		
		If `option` == `docs` == None, full application help and configuration is displayed:
		 - application name, version and license name
		 - application description and other documentation supplied with `Main.add_doc`
		 - a list of application options with their descriptions and current values
		 - a list of currently specified targets.
		
		If `option` is set to an option's name, only that option will be described.
		
		If `docs` is set to something else, only those docs will be displayed.
		"""
		
		...
	
	def parse_config(self):
		...
	
	def parse_cmdline(self, cmdline=None):
		cmdline = cmdline or sys.argv[1:]
		
		...
	
	def parse(self, cmdline=None):
		self.parse_config()
		self.parse_cmdline(cmdline)
